/*
 * GC.C
 * 1991-12-13
 *
 * Lee Daniel Crocker           lee@mport.com (uunet!mport!lee)
 * 5506 Camden Ave #D3          73407.2030@compuserve.com
 * San Jose, CA  95124
 *
 * A program to show, add, or delete comments in GIF streams.
 *
 * This code is released into the public domain completely free of
 * any copyrights and restrictions.  You are welcome (indeed
 * encouraged) to use, modify, and spread this code throughout the
 * known universe either freely or commercially.  Copy it, put your
 * name in the credits, and give it to your friends.  Put bugs in
 * it and give it to your enemies.  I don't care.  If you make
 * useful modifications, you are encouraged to send them to me for
 * future enhancement.
 *
 * Please note: this code is meant to be portable among most systems
 * which conform to ANSI X3.159-1989 with a few extensions not
 * provided by that standard.  In particular, a lot of the code may
 * like it does a lot of work for nothing to avoid assumptions about
 * the underlying character set.  Please think twice about changing
 * this code in the name of "simplification" or "optimization".
 */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>

/*
 * OS dependencies
 */

#if defined(MSDOS)
#define ADD_EXTENSIONS 1        /* Add ".gif", etc. to filenames */
#endif

#define GIFEXT  ".gif"  /* Leave in lowercase for Unix; make_filename */
#define TEXTEXT ".txt"

#define IOBUFSIZE   ((size_t)32768) /* Must be at least 1024 */

#define COMMA   0x2C    /* For portability to non-ASCII machines */
#define BANG    0x21
#define SEMICOLON 0x3B

#if defined(UNIX)
# include <unistd.h>        /* Need getpid() */
#elif defined(MSDOS)
# include <process.h>       /* Ditto */
#else
int
getpid(void)
{
    static int seeded = 0;

    if (!seeded) {
        srand((unsigned int)clock());
        seeded = 1;
    }
    return rand();
}
#endif

/*
 * Prototypes
 */

static char *basename(char *);
static void get_environment(char *);
static int usage(int);
static void temp_filename(char *, char *);
static void make_filename(char *, char *, char *);
static void fatal(int, char *, ...);
static void read_gif_header(char *);
static void write_gif_header(FILE *);
static void show_count(int, char *);
static void add_comment(char *, FILE *);
static void copy_file(FILE *, FILE *);
static void translate_char(int);
static void do_comment(FILE *);
static void skip_blocks(FILE *);
static void copy_blocks(FILE *, FILE *);
static void show_comments(FILE *);
static void delete_comments(FILE *, FILE *);

#define FATAL1(x) fatal(__LINE__,(x))
#define FATAL2(x,y) fatal(__LINE__,(x),(y))
#define FATAL3(x,y,z) fatal(__LINE__,(x),(y),(z))

/*
 * Globals
 */

static int verbose = 0, translate = 0;
static char command_name[FILENAME_MAX];     /* Made global for fatal() */
static FILE *gifin = NULL, *gifout = NULL;  /* Ditto */
static unsigned char *iobuf, *bufptr;       /* Shared by several functions */
static int s_count, a_count, d_count;
static int c_add, c_delete, c_show;         /* Commands */

/*
 * Strings
 */

static char Sgif[4] = { 0x47, 0x49, 0x46, 0x00 };   /* Must be ASCII */
static char S89a[4] = { 0x38, 0x39, 0x61, 0x00 };

static char Swrite_error[] =    "Write error on output.\n";
static char Sread_error[] =     "Read error from GIF file.\n";
static char Scannot_open[] =    "Cannot open \"%s\" for %s.\n";
static char Scannot_rename[] =  "Cannot rename \"%s\" to \"%s\".\n";
static char Sbad_data[] =       "Bad data in GIF stream.\n";
static char Soutput[] =         "output";
static char Sinput[] =          "input";

/*
 * Code begins here
 */

int
main(int argc, char *argv[])
{
    int c;
    char gifname[FILENAME_MAX]  = { '\0' };
    char textname[FILENAME_MAX] = { '\0' };
    char tname[2][FILENAME_MAX] = { '\0' };

    get_environment(*argv);
    if (argc == 1) return usage(1);

    s_count = a_count = d_count = 0;
    c_add = c_delete = 0;
    c_show = 1;

    while (--argc) {
        if (**++argv == '-') {      /* Command switch */
            while ((c = *++*argv) != '\0') {
                switch (tolower(c)) {
                case 'a':
                    c_add = 1;
                    c_show = 0;
                    break;
                case 'd':
                    c_delete = 1;
                    c_show = 0;
                    break;
                case 'h':
                case '?':
                    return usage(1);
                case 'v':
                    verbose = 1;
                    break;
                case 't':
                    translate = 1;
                    break;
                case 's':
                    c_show = 1;
                    c_add = c_delete = 0;
                    break;
                default:
                    return usage(0);
                }
            }
        } else {                    /* Filename on command line */
            if (!*gifname) {
                make_filename(gifname, *argv, GIFEXT);

                if ((iobuf = malloc(IOBUFSIZE)) == NULL) {
                    FATAL1("Cannot allocate I/O buffer.\n");
                }
                read_gif_header(gifname);

                if (!c_show) {
                    temp_filename(tname[0], gifname);

                    if ((gifout = fopen(tname[0], "wb")) == NULL) {
                        FATAL3(Scannot_open, "temporary", Soutput);
                    }
                    /*
                     * Signature must be 89a or later for comments.
                     */
                    if (c_add &&
                        strncmp((char*)iobuf+3, S89a, 3) < 0) {
                        strncpy((char*)iobuf+3, S89a, 3);
                    }
                    write_gif_header(gifout);
                }
            } else {
                if (!c_add) return usage(0);

                make_filename(textname, *argv, TEXTEXT);
                add_comment(textname, gifout);
            }
        }
    }
    if (gifin == NULL) return usage(0); /* Command but no GIF given */

    if (c_show) {
        show_comments(gifin);
        show_count(s_count, "found");
        fclose(gifin);
    } else {
        if (c_delete) {
            delete_comments(gifin, gifout);
            show_count(d_count, "deleted");
        }
        if (c_add) {
            show_count(a_count, "added");
        }
        copy_file(gifin, gifout);

        fclose(gifin);
        fclose(gifout);

        temp_filename(tname[1], gifname);

        if (rename(gifname, tname[1]) != 0) {
            FATAL3(Scannot_rename, gifname, tname[1]);
        }
        if (rename(tname[0], gifname) != 0) {
            FATAL3(Scannot_rename, tname[0], gifname);
        }
        remove(tname[1]);
    }
    return EXIT_SUCCESS;
}

/*
 * Get command name for Usage note.  This strips any path that may
 * have preceded the command, and strips and extension on MSDOS.
 * Pre-3.0 versions of MSDOS return the string "C" in argv[0].
 */

static void
get_environment(char *arg0)
{
    char *cp;

#ifdef MSDOS
    if (strcmp(arg0, "C") == 0) {
        strcpy(command_name, "gc");
    } else
#endif
    strcpy(command_name, basename(arg0));
#ifdef MSDOS
    if ((cp = strrchr(command_name, '.')) != NULL) {
        *cp = '\0';
    }
#endif
}

/*
 * Print usage note and optionally banner.  Banner is printed if
 * GC is invoked with no arguments or with -h or -?.
 */

static int
usage(int banner)
{
    if (banner) {
        fprintf(stderr, "%s: GIF commenting program\n"
                        "Version 1.01 1991-12-13\n"
                        "by Lee Daniel Crocker\n"
                        "73407.2030@compuserve.com\n\n", command_name);
    }
    fprintf(stderr, "Usage:\n"
                    "%s [-adsh?vt] giffile [textfile ...]\n\n"
                    "-a adds textfiles as comments,\n"
                    "-d deletes all comments,\n"
                    "-s (default) shows all comments.\n", command_name);
    fprintf(stderr, "-v sets verbose output mode.\n"
                    "-t sets ASCII character translation mode.\n"
                    "-h or -? prints this usage note.\n\n");
    fprintf(stderr, "The v and t flags can be combined with any command.\n"
                    "The a and d commands may be combined to delete all\n"
                    "comments after those added.\n"
                    "Text files only apply to -a command.\n");

    return EXIT_FAILURE;
}

/*
 * Return a pointer to the final element in a pathname.
 */

#ifdef MSDOS
static char pathseps[] = ":/\\";
#else
static char pathseps[] = "/";
#endif

static char *
basename(char *pathname)
{
    int i, len;
    char *cp, *rcp = pathname;

    len = (int)strlen(pathseps);
    for (i = 0; i < len; ++i) {
        if ((cp = strrchr(rcp, pathseps[i])) != NULL) rcp = cp+1;
    }
    return rcp;
}

/*
 * Fill the buffer tfname[] with the name of a temporary file in
 * the same directory as pathname[].  This is used instead of
 * the ANSI tempfile functions to ensure that the temporary is
 * in the same directory as the source file, and that if anything
 * goes wrong, it will not be deleted automatically.
 *
 * We use the process ID number to avoid collisions between
 * multiple instances of the same program running in the same
 * directory (for multitasking or network systems).
 */

static void
temp_filename(char *tfname, char *pathname)
{
    char *cp, temp_base[20];
    static int count = 0;

    sprintf(temp_base, "gc%d.%d", ((int)getpid() & 0xFFF), count++);

    strcpy(tfname, pathname);
    cp = basename(tfname);
    strcpy(cp, temp_base);
}

/*
 * Assemble filename from source and extension.  Some OSs may need
 * other tricks here (such as removing the version number from VMS
 * filenames).
 */

static void
make_filename(char *dest, char *src, char *ext)
{
    char *cp;

    strcpy(dest, src);
#ifdef ADD_EXTENSIONS
    if (strchr(basename(dest), '.') == NULL) {
        strcat(dest, ext);
    }
#endif
#ifdef MSDOS                        /* Unessential nicety for MSDOS */
    for (cp = dest; *cp != '\0'; ++cp) {
        *cp = (char)toupper(*cp);
    }
#endif
}

/*
 * Print fatal error message and exit.
 */

static void
fatal(int lineno, char *fmt, ...)
{
    va_list args;

    if (gifin) fclose(gifin);       /* Not strictly necessary, but just */
    if (gifout) fclose(gifout);     /* being paranoid */

    va_start(args, fmt);
    fprintf(stderr, "Error (line %d): ", lineno);
    vfprintf(stderr, fmt, args);
    va_end(args);

    exit(EXIT_FAILURE);
}

/*
 * Read global descriptor table and global colormap from GIF.
 * Print stats if in verbose mode.
 */

static void
read_gif_header(char *fname)
{
    int width, height, colors, crez;
    static char *crtext[]
        = { "8", "64", "512", "4096", "32k", "256k", "2M", "16M" };

    if ((gifin = fopen(fname, "rb")) == NULL) {
        FATAL3(Scannot_open, fname, Sinput);
    }
    if (fread(bufptr = iobuf, 1, 13, gifin) != 13) {
        FATAL1(Sread_error);
    } else {
        bufptr += 13;
    }
    if (strncmp(Sgif, (char*)iobuf, 3) != 0) {
        FATAL2("\"%s\" is not a GIF file.\n", fname);
    }
    if (iobuf[10] & 0x80) {
        colors = (1 << ((iobuf[10] & 0x07) + 1));
        if (fread(bufptr, 3, colors, gifin) != (size_t)colors) {
            FATAL1(Sread_error);
        } else {
            bufptr += (colors * 3);
        }
    } else {
        colors = 2;     /* No global colormap; assume B&W */
    }
    if (verbose) {
        width = (int)(iobuf[6]) + ((int)(iobuf[7]) << 8);
        height = (int)(iobuf[8]) + ((int)(iobuf[9]) << 8);
        crez = ((iobuf[10] >> 4) & 0x07);

        printf("%s: %d x %d, %d of %s colors\n",
            fname, width, height, colors, crtext[crez]);
    }
    /* Leave with (bufptr - iobuf) indicating bytes read */
}

/* Write the GIF header and global colormap just read back to
 * output file.
 */

static void
write_gif_header(FILE *outfile)
{
    size_t count = (size_t)(bufptr - iobuf);
    /*
     * Assume that input file is in same state as at the
     * end of read_gif_header().
     */
    if (fwrite(iobuf, 1, count, outfile) != count) {
        FATAL1(Swrite_error);
    }
}

/*
 * Tell the user what we've accomplished (with correct grammar).
 */

static void
show_count(int comment_count, char *action)
{
    char temp[12];

    if (verbose) {
        sprintf(temp, "%d", comment_count);
        printf("%s comment%s%s.\n",
            ((comment_count == 0) ? "No" : temp),
            ((comment_count == 1) ? " " : "s "), action);
    }
}

/*
 * We've read past the header, and have been given a filename to
 * add as a comment.  Write it out as a single GIF comment.
 * Sub-blocks are made as large as possible except for possibly
 * the last few bytes left in the buffer when the file is larger
 * than the i/o buffer.  Those are written as a small block before
 * another buffer is read in.
 */

static void
add_comment(char *tname, FILE *outfile)
{
    size_t count, bc;
    FILE *textin;
    static unsigned char extension_id[] = { BANG, 0xFE };

    if ((textin = fopen(tname, "rb")) == NULL) {
        fprintf(stderr, "Could not open \"%s\": skipping.\n", tname);
    }
    if (fwrite((char*)extension_id, 1, 2, outfile) != 2)
        FATAL1(Swrite_error);

    for (;;) {
        if ((count = fread(iobuf, 1, IOBUFSIZE, textin)) < 1) {
            break;
        } else {
            bufptr = iobuf;
            while (count) {
                bc = ((count > 255) ? 255 : count);
                putc(bc, outfile);
                if (fwrite(bufptr, 1, bc, outfile) != bc) {
                    FATAL1(Swrite_error);
                }
                if (verbose) {
                    fwrite(bufptr, 1, bc, stdout);
                }
                bufptr += bc;
                count -= bc;
            }
        }
    }
    putc(0, outfile);
    fclose(textin);
    ++a_count;
}

/*
 * We've been asked to add comments, so we don't need to parse the
 * rest of the GIF after we've done that, so just blast it across to
 * the output a buffer at a time.
 */

static void
copy_file(FILE *infile, FILE *outfile)
{
    size_t count;

    while ((count = fread(iobuf, 1, IOBUFSIZE, infile)) > 0) {
        if (fwrite(iobuf, 1, count, outfile) != count) {
            FATAL1(Swrite_error);
        }
    }
}

/*
 * Take ASCII char from GIF comment and show it readably on all
 * systems.  Alphanumerics and spaces are printed in the native
 * character set, and all other chars are shown in hex.
 *
 * Those of you who grew up with ASCII may be tempted to "optimize"
 * this, because it looks like a lot of work for nothing.  DON'T.
 * Others may want to make it more "portable" by replacing 0x0A
 * with '\n'.  DON'T.  This code is portable AS IS.  Trust me.
 */

static char *controls[] = { "NUL", "SOH", "STX", "ETX", "EOT",
    "ENQ", "ACK", "BEL", "BS", "HT", "LF", "VT", "FF", "CR",
    "SO", "SI", "DLE", "DC1", "DC2", "DC3", "DC4", "NAK", "SYN",
    "ETB", "CAN", "EM", "SUB", "ESC", "FS", "GS", "RS", "US" };

static char glyphs[] =  " !\"#$%&'()*+,-./0123456789:;<=>?"
                        "@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_"
                        "`abcdefghijklmnopqrstuvwxyz{|}~";

static void
translate_char(int c)
{
    if (c >= 0x20 && c <= 0x7E) {
        putchar(glyphs[c - 0x20]);
    } else if (c < 0x20) {
        printf("<%s>", controls[c]);
        /*
         * If we get an ASCII linefeed or formfeed, then we have
         * just printed <LF> or <FF>.  Now send a native newline
         * so that text output still looks somewhat normal.  This
         * is also how to tell a formfeed in the text from an 0xFF
         * byte--the former causes this newline.
         */
        if (c == 0x0A || c == 0x0C) {
            putchar('\n');
        }
    } else {
        printf("<%02x>", c);
    }
}

/*
 * Send text of comment to output.  Input is in ASCII,
 */

static void
do_comment(FILE *infile)
{
    size_t i, count;

    while ((count = getc(infile)) > 0) {
        if (fread(iobuf, 1, count, infile) != count) {
            FATAL1(Sread_error);
        }
        if (translate) {
            for (i = 0; i < count; ++i) {
                translate_char((int)iobuf[i]);
            }
        } else {
            fwrite(iobuf, 1, count, stdout);
        }
    }
}

/*
 * Read sub-blocks from GIF stream and discard.
 */

static void
skip_blocks(FILE *infile)
{
    size_t count;

    while ((count = getc(infile)) > 0) {
        if (fread(iobuf, 1, count, infile) != count) {
            FATAL1(Sread_error);
        }
    }
}

/*
 * Read sub-clocks from GIF stream and copy to output.
 */

static void
copy_blocks(FILE *infile, FILE *outfile)
{
    size_t count;

    while ((count = getc(infile)) > 0) {
        iobuf[0] = (unsigned char)count;
        if (fread(iobuf+1, 1, count, infile) != count) {
            FATAL1(Sread_error);
        }
        if (fwrite(iobuf, 1, count+1, outfile) != count+1) {
            FATAL1(Swrite_error);
        }
    }
    putc(0, outfile);
}

/*
 * Process GIF stream skipping everything but comments, which are
 * printed.  The whole stream must be read, because a comment may
 * be the last thing in a it.
 */

static void
show_comments(FILE *infile)
{
    int c;

    while ((c = getc(infile)) != EOF && c != SEMICOLON) {
        switch (c) {
        case COMMA:
            if (fread(iobuf, 1, 10, infile) != 10) {
                FATAL1(Sread_error);
            }
            skip_blocks(infile);
            break;
        case BANG:
            if ((c = getc(infile)) == 0xFE) {
                do_comment(infile);
                ++s_count;
            } else {
                skip_blocks(infile);
            }
            break;
        default:
            FATAL1(Sbad_data);
        }
    }
}

/*
 * Process GIF stream copying everything but comments to output.
 */

static void
delete_comments(FILE *infile, FILE *outfile)
{
    int c;

    while ((c = getc(infile)) != EOF && c != SEMICOLON) {
        switch (c) {
        case COMMA:
            *iobuf = COMMA;
            if (fread(iobuf+1, 1, 10, infile) != 10) {
                FATAL1(Sread_error);
            }
            if (fwrite(iobuf, 1, 11, outfile) != 11) {
                FATAL1(Swrite_error);
            }
            copy_blocks(infile, outfile);
            break;
        case BANG:
            if ((c = getc(infile)) == 0xFE) {
                skip_blocks(infile);
                ++d_count;
            } else {                /* Non-comment extension */
                putc(BANG, outfile);
                putc(c, outfile);
                copy_blocks(infile, outfile);
            }
            break;
        default:
            FATAL1(Sbad_data);
        }
    }
    if (c == SEMICOLON) {           /* Copy terminator */
        putc(c, outfile);
    }
}

